@@ -5,42 +5,28 @@
  * We do that by scanning the keyspace and for each pointer we have, we can try to
  * ask the allocator if moving it to a new address will help reduce fragmentation.
  *
- * Copyright (c) 2020, Redis Labs, Inc
+ * Copyright (c) 2020-Present, Redis Ltd.
  * All rights reserved.
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions are met:
- *
- *   * Redistributions of source code must retain the above copyright notice,
- *     this list of conditions and the following disclaimer.
- *   * Redistributions in binary form must reproduce the above copyright
- *     notice, this list of conditions and the following disclaimer in the
- *     documentation and/or other materials provided with the distribution.
- *   * Neither the name of Redis nor the names of its contributors may be used
- *     to endorse or promote products derived from this software without
- *     specific prior written permission.
- *
- * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
- * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
- * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
- * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
- * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
- * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
- * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
- * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
- * POSSIBILITY OF SUCH DAMAGE.
+ * Licensed under your choice of the Redis Source Available License 2.0
+ * (RSALv2) or the Server Side Public License v1 (SSPLv1).
  */
 
 #include "server.h"
-#include "cluster.h"
-#include <time.h>
-#include <assert.h>
 #include <stddef.h>
 
 #ifdef HAVE_DEFRAG
 
+typedef struct defragCtx {
+    void *privdata;
+    int slot;
+} defragCtx;
+
+typedef struct defragPubSubCtx {
+    kvstore *pubsub_channels;
+    dict *(*clientPubSubChannels)(client*);
+} defragPubSubCtx;
+
 /* this method was added to jemalloc in order to help us understand which
  * pointers are worthwhile moving and which aren't */
 int je_get_defrag_hint(void* ptr);
@@ -60,7 +46,7 @@ void* activeDefragAlloc(void *ptr) {
     /* move this allocation to a new allocation.
      * make sure not to use the thread cache. so that we don't get back the same
      * pointers we try to free */
-    size = zmalloc_size(ptr);
+    size = zmalloc_usable_size(ptr);
     newptr = zmalloc_no_tcache(size);
     memcpy(newptr, ptr, size);
     zfree_no_tcache(ptr);
@@ -84,14 +70,32 @@ sds activeDefragSds(sds sdsptr) {
     return NULL;
 }
 
-/* Defrag helper for robj and/or string objects
+/* Defrag helper for hfield strings
  *
  * returns NULL in case the allocation wasn't moved.
  * when it returns a non-null value, the old pointer was already released
  * and should NOT be accessed. */
-robj *activeDefragStringOb(robj* ob) {
+hfield activeDefragHfield(hfield hf) {
+    void *ptr = hfieldGetAllocPtr(hf);
+    void *newptr = activeDefragAlloc(ptr);
+    if (newptr) {
+        size_t offset = hf - (char*)ptr;
+        hf = (char*)newptr + offset;
+        return hf;
+    }
+    return NULL;
+}
+
+/* Defrag helper for robj and/or string objects with expected refcount.
+ *
+ * Like activeDefragStringOb, but it requires the caller to pass in the expected
+ * reference count. In some cases, the caller needs to update a robj whose
+ * reference count is not 1, in these cases, the caller must explicitly pass
+ * in the reference count, otherwise defragmentation will not be performed.
+ * Note that the caller is responsible for updating any other references to the robj. */
+robj *activeDefragStringObEx(robj* ob, int expected_refcount) {
     robj *ret = NULL;
-    if (ob->refcount!=1)
+    if (ob->refcount!=expected_refcount)
         return NULL;
 
     /* try to defrag robj (only if not an EMBSTR type (handled below). */
@@ -122,6 +126,15 @@ robj *activeDefragStringOb(robj* ob) {
     return ret;
 }
 
+/* Defrag helper for robj and/or string objects
+ *
+ * returns NULL in case the allocation wasn't moved.
+ * when it returns a non-null value, the old pointer was already released
+ * and should NOT be accessed. */
+robj *activeDefragStringOb(robj* ob) {
+    return activeDefragStringObEx(ob, 1);
+}
+
 /* Defrag helper for lua scripts
  *
  * returns NULL in case the allocation wasn't moved.
@@ -143,11 +156,20 @@ luaScript *activeDefragLuaScript(luaScript *script) {
 }
 
 /* Defrag helper for dict main allocations (dict struct, and hash tables).
- * receives a pointer to the dict* and implicitly updates it when the dict
- * struct itself was moved. Returns a stat of how many pointers were moved. */
-void dictDefragTables(dict* d) {
+ * Receives a pointer to the dict* and return a new dict* when the dict
+ * struct itself was moved.
+ * 
+ * Returns NULL in case the allocation wasn't moved.
+ * When it returns a non-null value, the old pointer was already released
+ * and should NOT be accessed. */
+dict *dictDefragTables(dict *d) {
+    dict *ret = NULL;
     dictEntry **newtable;
+    /* handle the dict struct */
+    if ((ret = activeDefragAlloc(d)))
+        d = ret;
     /* handle the first hash table */
+    if (!d->ht_table[0]) return ret; /* created but unused */
     newtable = activeDefragAlloc(d->ht_table[0]);
     if (newtable)
         d->ht_table[0] = newtable;
@@ -157,6 +179,7 @@ void dictDefragTables(dict* d) {
         if (newtable)
             d->ht_table[1] = newtable;
     }
+    return ret;
 }
 
 /* Internal function used by zslDefrag */
@@ -243,6 +266,31 @@ void activeDefragSdsDictCallback(void *privdata, const dictEntry *de) {
     UNUSED(de);
 }
 
+void activeDefragHfieldDictCallback(void *privdata, const dictEntry *de) {
+    dict *d = privdata;
+    hfield newhf, hf = dictGetKey(de);
+
+    if (hfieldGetExpireTime(hf) == EB_EXPIRE_TIME_INVALID) {
+        /* If the hfield does not have TTL, we directly defrag it. */
+        newhf = activeDefragHfield(hf);
+    } else {
+        /* Update its reference in the ebucket while defragging it. */
+        ebuckets *eb = hashTypeGetDictMetaHFE(d);
+        newhf = ebDefragItem(eb, &hashFieldExpireBucketsType, hf, (ebDefragFunction *)activeDefragHfield);
+    }
+    if (newhf) {
+        /* We can't search in dict for that key after we've released
+         * the pointer it holds, since it won't be able to do the string
+         * compare, but we can find the entry using key hash and pointer. */
+        dictUseStoredKeyApi(d, 1);
+        uint64_t hash = dictGetHash(d, newhf);
+        dictUseStoredKeyApi(d, 0);
+        dictEntry *de = dictFindEntryByPtrAndHash(d, hf, hash);
+        serverAssert(de);
+        dictSetKey(d, de, newhf);
+    }
+}
+
 /* Defrag a dict with sds key and optional value (either ptr, sds or robj string) */
 void activeDefragSdsDict(dict* d, int val_type) {
     unsigned long cursor = 0;
@@ -261,6 +309,20 @@ void activeDefragSdsDict(dict* d, int val_type) {
     } while (cursor != 0);
 }
 
+/* Defrag a dict with hfield key and sds value. */
+void activeDefragHfieldDict(dict *d) {
+    unsigned long cursor = 0;
+    dictDefragFunctions defragfns = {
+        .defragAlloc = activeDefragAlloc,
+        .defragKey = NULL, /* Will be defragmented in activeDefragHfieldDictCallback. */
+        .defragVal = (dictDefragAllocFunction *)activeDefragSds
+    };
+    do {
+        cursor = dictScanDefrag(d, cursor, activeDefragHfieldDictCallback,
+                                &defragfns, d);
+    } while (cursor != 0);
+}
+
 /* Defrag a list of ptr, sds or robj string values */
 void activeDefragList(list *l, int val_type) {
     listNode *ln, *newln;
@@ -415,10 +477,10 @@ void scanLaterHash(robj *ob, unsigned long *cursor) {
     dict *d = ob->ptr;
     dictDefragFunctions defragfns = {
         .defragAlloc = activeDefragAlloc,
-        .defragKey = (dictDefragAllocFunction *)activeDefragSds,
+        .defragKey = NULL, /* Will be defragmented in activeDefragHfieldDictCallback. */
         .defragVal = (dictDefragAllocFunction *)activeDefragSds
     };
-    *cursor = dictScanDefrag(d, *cursor, scanCallbackCountScanned, &defragfns, NULL);
+    *cursor = dictScanDefrag(d, *cursor, activeDefragHfieldDictCallback, &defragfns, d);
 }
 
 void defragQuicklist(redisDb *db, dictEntry *kde) {
@@ -457,11 +519,9 @@ void defragZsetSkiplist(redisDb *db, dictEntry *kde) {
         }
         dictReleaseIterator(di);
     }
-    /* handle the dict struct */
-    if ((newdict = activeDefragAlloc(zs->dict)))
+    /* defrag the dict struct and tables */
+    if ((newdict = dictDefragTables(zs->dict)))
         zs->dict = newdict;
-    /* defrag the dict tables */
-    dictDefragTables(zs->dict);
 }
 
 void defragHash(redisDb *db, dictEntry *kde) {
@@ -472,12 +532,10 @@ void defragHash(redisDb *db, dictEntry *kde) {
     if (dictSize(d) > server.active_defrag_max_scan_fields)
         defragLater(db, kde);
     else
-        activeDefragSdsDict(d, DEFRAG_SDS_DICT_VAL_IS_SDS);
-    /* handle the dict struct */
-    if ((newd = activeDefragAlloc(ob->ptr)))
+        activeDefragHfieldDict(d);
+    /* defrag the dict struct and tables */
+    if ((newd = dictDefragTables(ob->ptr)))
         ob->ptr = newd;
-    /* defrag the dict tables */
-    dictDefragTables(ob->ptr);
 }
 
 void defragSet(redisDb *db, dictEntry *kde) {
@@ -489,11 +547,9 @@ void defragSet(redisDb *db, dictEntry *kde) {
         defragLater(db, kde);
     else
         activeDefragSdsDict(d, DEFRAG_SDS_DICT_NO_VAL);
-    /* handle the dict struct */
-    if ((newd = activeDefragAlloc(ob->ptr)))
+    /* defrag the dict struct and tables */
+    if ((newd = dictDefragTables(ob->ptr)))
         ob->ptr = newd;
-    /* defrag the dict tables */
-    dictDefragTables(ob->ptr);
 }
 
 /* Defrag callback for radix tree iterator, called for each node,
@@ -669,32 +725,43 @@ void defragModule(redisDb *db, dictEntry *kde) {
 }
 
 /* for each key we scan in the main dict, this function will attempt to defrag
- * all the various pointers it has. Returns a stat of how many pointers were
- * moved. */
-void defragKey(redisDb *db, dictEntry *de) {
+ * all the various pointers it has. */
+void defragKey(defragCtx *ctx, dictEntry *de) {
     sds keysds = dictGetKey(de);
-    robj *newob, *ob;
+    robj *newob, *ob = dictGetVal(de);
     unsigned char *newzl;
     sds newsds;
-
+    redisDb *db = ctx->privdata;
+    int slot = ctx->slot;
     /* Try to defrag the key name. */
     newsds = activeDefragSds(keysds);
     if (newsds) {
-        dictSetKey(db->dict, de, newsds);
-        if (dictSize(db->expires)) {
+        kvstoreDictSetKey(db->keys, slot, de, newsds);
+        if (kvstoreDictSize(db->expires, slot)) {
             /* We can't search in db->expires for that key after we've released
              * the pointer it holds, since it won't be able to do the string
              * compare, but we can find the entry using key hash and pointer. */
-            uint64_t hash = dictGetHash(db->dict, newsds);
-            dictEntry *expire_de = dictFindEntryByPtrAndHash(db->expires, keysds, hash);
-            if (expire_de) dictSetKey(db->expires, expire_de, newsds);
+            uint64_t hash = kvstoreGetHash(db->expires, newsds);
+            dictEntry *expire_de = kvstoreDictFindEntryByPtrAndHash(db->expires, slot, keysds, hash);
+            if (expire_de) kvstoreDictSetKey(db->expires, slot, expire_de, newsds);
         }
+
+        /* Update the key's reference in the dict's metadata or the listpackEx. */
+        if (unlikely(ob->type == OBJ_HASH))
+            hashTypeUpdateKeyRef(ob, newsds);
     }
 
     /* Try to defrag robj and / or string value. */
-    ob = dictGetVal(de);
-    if ((newob = activeDefragStringOb(ob))) {
-        dictSetVal(db->dict, de, newob);
+    if (unlikely(ob->type == OBJ_HASH && hashTypeGetMinExpire(ob, 0) != EB_EXPIRE_TIME_INVALID)) {
+        /* Update its reference in the ebucket while defragging it. */
+        newob = ebDefragItem(&db->hexpires, &hashExpireBucketsType, ob,
+                             (ebDefragFunction *)activeDefragStringOb);
+    } else {
+        /* If the dict doesn't have metadata, we directly defrag it. */
+        newob = activeDefragStringOb(ob);
+    }
+    if (newob) {
+        kvstoreDictSetVal(db->keys, slot, de, newob);
         ob = newob;
     }
 
@@ -734,6 +801,12 @@ void defragKey(redisDb *db, dictEntry *de) {
         if (ob->encoding == OBJ_ENCODING_LISTPACK) {
             if ((newzl = activeDefragAlloc(ob->ptr)))
                 ob->ptr = newzl;
+        } else if (ob->encoding == OBJ_ENCODING_LISTPACK_EX) {
+            listpackEx *newlpt, *lpt = (listpackEx*)ob->ptr;
+            if ((newlpt = activeDefragAlloc(lpt)))
+                ob->ptr = lpt = newlpt;
+            if ((newzl = activeDefragAlloc(lpt->lp)))
+                lpt->lp = newzl;
         } else if (ob->encoding == OBJ_ENCODING_HT) {
             defragHash(db, de);
         } else {
@@ -751,7 +824,7 @@ void defragKey(redisDb *db, dictEntry *de) {
 /* Defrag scan callback for the main db dictionary. */
 void defragScanCallback(void *privdata, const dictEntry *de) {
     long long hits_before = server.stat_active_defrag_hits;
-    defragKey((redisDb*)privdata, (dictEntry*)de);
+    defragKey((defragCtx*)privdata, (dictEntry*)de);
     if (server.stat_active_defrag_hits != hits_before)
         server.stat_active_defrag_key_hits++;
     else
@@ -766,20 +839,68 @@ void defragScanCallback(void *privdata, const dictEntry *de) {
  * or not, a false detection can cause the defragmenter to waste a lot of CPU
  * without the possibility of getting any results. */
 float getAllocatorFragmentation(size_t *out_frag_bytes) {
-    size_t resident, active, allocated;
-    zmalloc_get_allocator_info(&allocated, &active, &resident);
-    float frag_pct = ((float)active / allocated)*100 - 100;
-    size_t frag_bytes = active - allocated;
+    size_t resident, active, allocated, frag_smallbins_bytes;
+    zmalloc_get_allocator_info(1, &allocated, &active, &resident, NULL, NULL, &frag_smallbins_bytes);
+
+    if (server.lua_arena != UINT_MAX) {
+        size_t lua_resident, lua_active, lua_allocated, lua_frag_smallbins_bytes;
+        zmalloc_get_allocator_info_by_arena(server.lua_arena, 0, &lua_allocated, &lua_active, &lua_resident, &lua_frag_smallbins_bytes);
+        resident -= lua_resident;
+        active -= lua_active;
+        allocated -= lua_allocated;
+        frag_smallbins_bytes -= lua_frag_smallbins_bytes;
+    }
+
+    /* Calculate the fragmentation ratio as the proportion of wasted memory in small
+     * bins (which are defraggable) relative to the total allocated memory (including large bins).
+     * This is because otherwise, if most of the memory usage is large bins, we may show high percentage,
+     * despite the fact it's not a lot of memory for the user. */
+    float frag_pct = (float)frag_smallbins_bytes / allocated * 100;
     float rss_pct = ((float)resident / allocated)*100 - 100;
     size_t rss_bytes = resident - allocated;
     if(out_frag_bytes)
-        *out_frag_bytes = frag_bytes;
+        *out_frag_bytes = frag_smallbins_bytes;
     serverLog(LL_DEBUG,
-        "allocated=%zu, active=%zu, resident=%zu, frag=%.0f%% (%.0f%% rss), frag_bytes=%zu (%zu rss)",
-        allocated, active, resident, frag_pct, rss_pct, frag_bytes, rss_bytes);
+        "allocated=%zu, active=%zu, resident=%zu, frag=%.2f%% (%.2f%% rss), frag_bytes=%zu (%zu rss)",
+        allocated, active, resident, frag_pct, rss_pct, frag_smallbins_bytes, rss_bytes);
     return frag_pct;
 }
 
+/* Defrag scan callback for the pubsub dictionary. */
+void defragPubsubScanCallback(void *privdata, const dictEntry *de) {
+    defragCtx *ctx = privdata;
+    defragPubSubCtx *pubsub_ctx = ctx->privdata;
+    kvstore *pubsub_channels = pubsub_ctx->pubsub_channels;
+    robj *newchannel, *channel = dictGetKey(de);
+    dict *newclients, *clients = dictGetVal(de);
+
+    /* Try to defrag the channel name. */
+    serverAssert(channel->refcount == (int)dictSize(clients) + 1);
+    newchannel = activeDefragStringObEx(channel, dictSize(clients) + 1);
+    if (newchannel) {
+        kvstoreDictSetKey(pubsub_channels, ctx->slot, (dictEntry*)de, newchannel);
+
+        /* The channel name is shared by the client's pubsub(shard) and server's
+         * pubsub(shard), after defraging the channel name, we need to update
+         * the reference in the clients' dictionary. */
+        dictIterator *di = dictGetIterator(clients);
+        dictEntry *clientde;
+        while((clientde = dictNext(di)) != NULL) {
+            client *c = dictGetKey(clientde);
+            dictEntry *pubsub_channel = dictFind(pubsub_ctx->clientPubSubChannels(c), newchannel);
+            serverAssert(pubsub_channel);
+            dictSetKey(pubsub_ctx->clientPubSubChannels(c), pubsub_channel, newchannel);
+        }
+        dictReleaseIterator(di);
+    }
+
+    /* Try to defrag the dictionary of clients that is stored as the value part. */
+    if ((newclients = dictDefragTables(clients)))
+        kvstoreDictSetVal(pubsub_channels, ctx->slot, (dictEntry*)de, newclients);
+
+    server.stat_active_defrag_scanned++;
+}
+
 /* We may need to defrag other globals, one small allocation can hold a full allocator run.
  * so although small, it is still important to defrag these */
 void defragOtherGlobals(void) {
@@ -789,6 +910,8 @@ void defragOtherGlobals(void) {
      * that remain static for a long time */
     activeDefragSdsDict(evalScriptsDict(), DEFRAG_SDS_DICT_VAL_LUA_SCRIPT);
     moduleDefragGlobals();
+    kvstoreDictLUTDefrag(server.pubsub_channels, dictDefragTables);
+    kvstoreDictLUTDefrag(server.pubsubshard_channels, dictDefragTables);
 }
 
 /* returns 0 more work may or may not be needed (see non-zero cursor),
@@ -824,7 +947,7 @@ static sds defrag_later_current_key = NULL;
 static unsigned long defrag_later_cursor = 0;
 
 /* returns 0 if no more work needs to be been done, and 1 if time is up and more work is needed. */
-int defragLaterStep(redisDb *db, long long endtime) {
+int defragLaterStep(redisDb *db, int slot, long long endtime) {
     unsigned int iterations = 0;
     unsigned long long prev_defragged = server.stat_active_defrag_hits;
     unsigned long long prev_scanned = server.stat_active_defrag_scanned;
@@ -854,7 +977,7 @@ int defragLaterStep(redisDb *db, long long endtime) {
         }
 
         /* each time we enter this function we need to fetch the key from the dict again (if it still exists) */
-        dictEntry *de = dictFind(db->dict, defrag_later_current_key);
+        dictEntry *de = kvstoreDictFind(db->keys, slot, defrag_later_current_key);
         key_defragged = server.stat_active_defrag_hits;
         do {
             int quit = 0;
@@ -899,7 +1022,8 @@ void computeDefragCycles(void) {
             return;
     }
 
-    /* Calculate the adaptive aggressiveness of the defrag */
+    /* Calculate the adaptive aggressiveness of the defrag based on the current
+     * fragmentation and configurations. */
     int cpu_pct = INTERPOLATE(frag_pct,
             server.active_defrag_threshold_lower,
             server.active_defrag_threshold_upper,
@@ -908,10 +1032,15 @@ void computeDefragCycles(void) {
     cpu_pct = LIMIT(cpu_pct,
             server.active_defrag_cycle_min,
             server.active_defrag_cycle_max);
-     /* We allow increasing the aggressiveness during a scan, but don't
-      * reduce it. */
-    if (cpu_pct > server.active_defrag_running) {
+
+    /* Normally we allow increasing the aggressiveness during a scan, but don't
+     * reduce it, since we should not lower the aggressiveness when fragmentation
+     * drops. But when a configuration is made, we should reconsider it. */
+    if (cpu_pct > server.active_defrag_running ||
+        server.active_defrag_configuration_changed)
+    {
         server.active_defrag_running = cpu_pct;
+        server.active_defrag_configuration_changed = 0;
         serverLog(LL_VERBOSE,
             "Starting active defrag, frag=%.0f%%, frag_bytes=%zu, cpu=%d%%",
             frag_pct, frag_bytes, cpu_pct);
@@ -922,9 +1051,11 @@ void computeDefragCycles(void) {
  * This works in a similar way to activeExpireCycle, in the sense that
  * we do incremental work across calls. */
 void activeDefragCycle(void) {
+    static int slot = -1;
     static int current_db = -1;
-    static unsigned long cursor = 0;
-    static unsigned long expires_cursor = 0;
+    static int defrag_later_item_in_progress = 0;
+    static int defrag_stage = 0;
+    static unsigned long defrag_cursor = 0;
     static redisDb *db = NULL;
     static long long start_scan, start_stat;
     unsigned int iterations = 0;
@@ -932,19 +1063,23 @@ void activeDefragCycle(void) {
     unsigned long long prev_scanned = server.stat_active_defrag_scanned;
     long long start, timelimit, endtime;
     mstime_t latency;
+    int all_stages_finished = 0;
     int quit = 0;
 
     if (!server.active_defrag_enabled) {
         if (server.active_defrag_running) {
             /* if active defrag was disabled mid-run, start from fresh next time. */
             server.active_defrag_running = 0;
+            server.active_defrag_configuration_changed = 0;
             if (db)
                 listEmpty(db->defrag_later);
             defrag_later_current_key = NULL;
             defrag_later_cursor = 0;
             current_db = -1;
-            cursor = 0;
-            expires_cursor = 0;
+            defrag_stage = 0;
+            defrag_cursor = 0;
+            slot = -1;
+            defrag_later_item_in_progress = 0;
             db = NULL;
             goto update_metrics;
         }
@@ -959,6 +1094,14 @@ void activeDefragCycle(void) {
     run_with_period(1000) {
         computeDefragCycles();
     }
+
+    /* Normally it is checked once a second, but when there is a configuration
+     * change, we want to check it as soon as possible. */
+    if (server.active_defrag_configuration_changed) {
+        computeDefragCycles();
+        server.active_defrag_configuration_changed = 0;
+    }
+
     if (!server.active_defrag_running)
         return;
 
@@ -972,9 +1115,9 @@ void activeDefragCycle(void) {
     dictDefragFunctions defragfns = {.defragAlloc = activeDefragAlloc};
     do {
         /* if we're not continuing a scan from the last call or loop, start a new one */
-        if (!cursor && !expires_cursor) {
+        if (!defrag_stage && !defrag_cursor && (slot < 0)) {
             /* finish any leftovers from previous db before moving to the next one */
-            if (db && defragLaterStep(db, endtime)) {
+            if (db && defragLaterStep(db, slot, endtime)) {
                 quit = 1; /* time is up, we didn't finish all the work */
                 break; /* this will exit the function and we'll continue on the next cycle */
             }
@@ -993,7 +1136,10 @@ void activeDefragCycle(void) {
 
                 start_scan = now;
                 current_db = -1;
-                cursor = 0;
+                defrag_stage = 0;
+                defrag_cursor = 0;
+                slot = -1;
+                defrag_later_item_in_progress = 0;
                 db = NULL;
                 server.active_defrag_running = 0;
 
@@ -1009,38 +1155,79 @@ void activeDefragCycle(void) {
             }
 
             db = &server.db[current_db];
-            cursor = 0;
+            kvstoreDictLUTDefrag(db->keys, dictDefragTables);
+            kvstoreDictLUTDefrag(db->expires, dictDefragTables);
+            defrag_stage = 0;
+            defrag_cursor = 0;
+            slot = -1;
+            defrag_later_item_in_progress = 0;
         }
 
+        /* This array of structures holds the parameters for all defragmentation stages. */
+        typedef struct defragStage {
+            kvstore *kvs;
+            dictScanFunction *scanfn;
+            void *privdata;
+        } defragStage;
+        defragStage defrag_stages[] = {
+            {db->keys, defragScanCallback, db},
+            {db->expires, scanCallbackCountScanned, NULL},
+            {server.pubsub_channels, defragPubsubScanCallback,
+                &(defragPubSubCtx){server.pubsub_channels, getClientPubSubChannels}},
+            {server.pubsubshard_channels, defragPubsubScanCallback,
+                &(defragPubSubCtx){server.pubsubshard_channels, getClientPubSubShardChannels}},
+        };
         do {
+            int num_stages = sizeof(defrag_stages) / sizeof(defrag_stages[0]);
+            serverAssert(defrag_stage < num_stages);
+            defragStage *current_stage = &defrag_stages[defrag_stage];
+
             /* before scanning the next bucket, see if we have big keys left from the previous bucket to scan */
-            if (defragLaterStep(db, endtime)) {
+            if (defragLaterStep(db, slot, endtime)) {
                 quit = 1; /* time is up, we didn't finish all the work */
                 break; /* this will exit the function and we'll continue on the next cycle */
             }
 
-            /* Scan the keyspace dict unless we're scanning the expire dict. */
-            if (!expires_cursor)
-                cursor = dictScanDefrag(db->dict, cursor, defragScanCallback,
-                                        &defragfns, db);
+            if (!defrag_later_item_in_progress) {
+                /* Continue defragmentation from the previous stage.
+                 * If slot is -1, it means this stage starts from the first non-empty slot. */
+                if (slot == -1) slot = kvstoreGetFirstNonEmptyDictIndex(current_stage->kvs);
+                defrag_cursor = kvstoreDictScanDefrag(current_stage->kvs, slot, defrag_cursor,
+                    current_stage->scanfn, &defragfns, &(defragCtx){current_stage->privdata, slot});
+            }
 
-            /* When done scanning the keyspace dict, we scan the expire dict. */
-            if (!cursor)
-                expires_cursor = dictScanDefrag(db->expires, expires_cursor,
-                                                scanCallbackCountScanned,
-                                                &defragfns, NULL);
+            if (!defrag_cursor) {
+                /* Move to the next slot only if regular and large item scanning has been completed. */
+                if (listLength(db->defrag_later) > 0) {
+                    defrag_later_item_in_progress = 1;
+                    continue;
+                }
 
+                /* Move to the next slot in the current stage. If we've reached the end, move to the next stage. */
+                if ((slot = kvstoreGetNextNonEmptyDictIndex(current_stage->kvs, slot)) == -1)
+                    defrag_stage++;
+                defrag_later_item_in_progress = 0;
+            }
+
+            /* Check if all defragmentation stages have been processed.
+             * If so, mark as finished and reset the stage counter to move on to next database. */
+            if (defrag_stage == num_stages) {
+                all_stages_finished = 1;
+                defrag_stage = 0;
+            }
+    
             /* Once in 16 scan iterations, 512 pointer reallocations. or 64 keys
              * (if we have a lot of pointers in one hash bucket or rehashing),
              * check if we reached the time limit.
              * But regardless, don't start a new db in this loop, this is because after
              * the last db we call defragOtherGlobals, which must be done in one cycle */
-            if (!(cursor || expires_cursor) ||
+            if (all_stages_finished ||
                 ++iterations > 16 ||
                 server.stat_active_defrag_hits - prev_defragged > 512 ||
                 server.stat_active_defrag_scanned - prev_scanned > 64)
             {
-                if (!cursor || ustime() > endtime) {
+                /* Quit if all stages were finished or timeout. */
+                if (all_stages_finished || ustime() > endtime) {
                     quit = 1;
                     break;
                 }
@@ -1048,7 +1235,7 @@ void activeDefragCycle(void) {
                 prev_defragged = server.stat_active_defrag_hits;
                 prev_scanned = server.stat_active_defrag_scanned;
             }
-        } while((cursor || expires_cursor) && !quit);
+        } while(!all_stages_finished && !quit);
     } while(!quit);
 
     latencyEndMonitor(latency);
